
""" The :mod:`dubious.pory` module defines the main class that will act as the bot: :class:`Pory`.

    It also includes a special subclass, :class:`ConfiguredPory`, that makes per-guild (read: per-server) configuration easier. """


from __future__ import annotations

import dataclasses as dc
import json
import multiprocessing as mp
import traceback as tb
import types
import typing as t
import pprint

import flask
from nacl.exceptions import BadSignatureError
from nacl.signing import VerifyKey

from dubious import callback as cb
from dubious.discord import api, disc, req


class CommandFailure(Exception):
    """ This class is a front-facing :class:`Exception` class to be thrown when an error is made by a user of the bot.
        
        It will be caught by :class:`Pory` and its :func:`~CommandFailure.error_embed` method will be used to decorate an appropriate error message.
        
        Can be subclassed to alter the :func:`~CommandFailure.error_embed` method. """

    message: str
    def __init__(self, message: str) -> None:
        super().__init__()
        self.message = message

    def error_embed(self):
        """ Returns an :class:`.Embed` object which will be sent as the front-facing error message. """
        return api.Embed(color=0xFF3300, description=str(self.message))

def _match_response_type(response: api.InteractionCallbackData | None, ixn_type: api.InteractionType) -> api.InteractionCallbackType:
    match response:
        case api.ResponseMessage():
            return api.InteractionCallbackType.CHANNEL_MESSAGE_WITH_SOURCE
        case api.ResponseModal():
            return api.InteractionCallbackType.MODAL
    match ixn_type:
        case api.InteractionType.PING:
            return api.InteractionCallbackType.PONG
        case api.InteractionType.APPLICATION_COMMAND | api.InteractionType.MODAL_SUBMIT:
            return api.InteractionCallbackType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE
        case api.InteractionType.MESSAGE_COMPONENT:
            return api.InteractionCallbackType.DEFERRED_UPDATE_MESSAGE
    raise Exception()

@dc.dataclass
class Pory:
    """ This class defines bot functionality.

        Once an instance is created, commands can be added using its :attr:`.on_command` :class:`.CommandGroup` as a decorator around the desired callback function.

        Also contains various utility functions for things like sending messages to specific channels and sending followup responses to interactions. """

    app_id: dc.InitVar[str]
    """ The application ID shown in the ``General Information`` section of the `application's settings <https://discord.com/developers/applications>`_. This value is converted into a :class:`.Snowflake` object and stored in :attr:`.id`. """
    public_key: str
    """ The public key shown in the ``General Information`` section of the `application's settings <https://discord.com/developers/applications>`_. """
    token: str
    """ The token generated by the *Reset Token* option in the ``Bot`` section of the `application's settings <https://discord.com/developers/applications>`_. """

    id: disc.Snowflake = dc.field(init=False)
    """ The application ID for the bot as a :class:`.Snowflake`."""

    def __post_init__(self, app_id: str):
        self.id = disc.Snowflake(app_id)

    on_command: cb.CommandGroup = dc.field(default_factory=cb.CommandGroup.new, init=False)
    """ :class:`.CommandGroup` for callbacks to register as `Slash Commands <https://discord.com/developers/docs/interactions/application-commands#slash-commands>`_. """

    on_message: cb.CommandGroup = dc.field(default_factory=cb.CommandGroup.new, init=False)
    """ :class:`.CommandGroup` for callbacks to register as `Message Commands <https://discord.com/developers/docs/interactions/application-commands#message-commands>`_. """

    on_user: cb.CommandGroup = dc.field(default_factory=cb.CommandGroup.new, init=False)
    """ :class:`.CommandGroup` for callbacks to register as `User Commands <https://discord.com/developers/docs/interactions/application-commands#user-commands>`_. """

    on_modal: cb.CallbackGroup = dc.field(default_factory=cb.CallbackGroup, init=False)
    """ :class:`.CallbackGroup` for internal callbacks that corespond to Modal (popup) interactions. """

    on_component: cb.CallbackGroup = dc.field(default_factory=cb.CallbackGroup, init=False)
    """ :class:`.CallbackGroup` for internal callbacks that corespond to `Message Component interactions <https://discord.com/developers/docs/interactions/message-components>`_. """
    
    def flask(self):
        """ The entry point for the app. Starts a :class:`~flask.Flask` instance and accepts ``POST`` requests to the ``/interactions`` endpoint. """

        app = flask.Flask(__name__)
        
        self._sync_commands()
        verify_key = VerifyKey(bytes.fromhex(self.public_key))

        @app.route("/")
        def index():
            return "why are you here"

        @app.route("/interactions", methods=["POST"])
        def _():
            signature = flask.request.headers["X-Signature-Ed25519"]
            timestamp = flask.request.headers["X-Signature-Timestamp"]
            body = flask.request.data.decode("utf-8")

            try:
                verify_key.verify(f"{timestamp}{body}".encode(), bytes.fromhex(signature))
            except BadSignatureError:
                flask.abort(401, "Invalid request signature")

            if not flask.request.json:
                return flask.abort(405, "Bad request")
            ixn_json = flask.request.json
            ixn = disc.cast(api.Interaction, ixn_json)
            res = self._handle(ixn)
            return flask.jsonify(dc.asdict(res))

        return app

    def _sync_commands(self):
        """ Collects the bot's Discord-registered commands with an API request and compares each of them with the commands defined in code with the bot, updating as needed. """

        # Collect local commands
        prepared_on_command = self.on_command.get_commands(as_type=api.ApplicationCommandType.CHAT_INPUT)
        prepared_on_message = self.on_message.get_commands(as_type=api.ApplicationCommandType.MESSAGE)
        prepared_on_user = self.on_user.get_commands(as_type=api.ApplicationCommandType.USER)

        # Collect registered commands
        registered_commands: list[api.ApplicationCommand] = []
        guilds = req.GetCurrentUserGuilds().do_with(self.token)
        for guild in guilds:
            registered_commands += req.GetGuildApplicationCommands(self.id, guild.id).do_with(self.token)
        registered_commands += req.GetGlobalApplicationCommands(self.id).do_with(self.token)

        # Compare
        for registered in registered_commands:
            match registered.type:
                case api.ApplicationCommandType.CHAT_INPUT:
                    prepared = prepared_on_command
                case api.ApplicationCommandType.MESSAGE:
                    prepared = prepared_on_message
                case api.ApplicationCommandType.USER:
                    prepared = prepared_on_user
                case _:
                    raise Exception(f"While syncing commands, encountered an unknown ApplicationCommandType: {registered.type}")
            self._sync_if_exists(registered, prepared)

        # Create local commands in Discord if they weren't found
        for to_register in (
            list(prepared_on_command.values()) +
            list(prepared_on_message.values()) +
            list(prepared_on_user.values())
        ):
            self._create_command(to_register)

    def _sync_if_exists(self, registered: api.ApplicationCommand, prepared: dict[str, cb.Command]):
        """ Compares a pre-existing :class:`.api.ApplicationCommand` retrieved from the Discord API, to a collection of local yet-to-be-registered :class:`.Command`s.
        
            If the pre-existing command matches one of the local commands, check if they match, and if they don't, update the command with an API request.
            
            If the pre-existing command doesn't match, delete the command with an API request.
            
            This function alters the ``prepared`` argument, popping items which have been found. """

        if not registered.name in prepared:
            if not registered.guild_id:
                print(f"Deleting registered command: {registered.name}")
                req.DeleteGlobalApplicationCommand(self.id, registered.id).do_with(self.token)
            else:
                print(f"Deleting registered command in guild {registered.guild_id}: {registered.name}")
                req.DeleteGuildApplicationCommand(self.id, registered.guild_id, registered.id).do_with(self.token)

        elif registered.guild_id == prepared[registered.name].guild_id:
            if not prepared[registered.name]._compare_with(registered):
                if not registered.guild_id:
                    print(f"Updating registered command: {registered.name}")
                    req.EditGlobalApplicationCommand(self.id, registered.id, disc.cast(req.EditGlobalApplicationCommand.Form, registered))
                else:
                    print(f"Updating registered command in guild {registered.guild_id}: {registered.name}")
                    req.EditGuildApplicationCommand(self.id, registered.guild_id, registered.id, disc.cast(req.EditGuildApplicationCommand.Form, registered))
            prepared.pop(registered.name)

    def _create_command(self, to_register: cb.Command):
        """ Sends :class:`dubious.discord.req.CreateGlobalApplicationCommand` requests to the Discord API to register a new command. """

        if not to_register.guild_id:
            print(f"Creating new command: {to_register.name}")
            req.CreateGlobalApplicationCommand(self.id, form=disc.cast(req.CreateGlobalApplicationCommand.Form, to_register)).do_with(self.token)
        else:
            print(f"Creating new command in guild {to_register.guild_id}: {to_register.name}")
            req.CreateGuildApplicationCommand(self.id, to_register.guild_id, form=disc.cast(req.CreateGuildApplicationCommand.Form, to_register)).do_with(self.token)

    def _handle(self, ixn: api.Interaction):
        """ Handles all incoming interaction POSTs to the Flask server, after their data have been cast to an :class:`.Interaction`.
            
            For :class:`.Interaction`\\ s that match a callback defined in the :class:`Pory`'s callbacks, it gets the coresponding callback and runs it. """

        match ixn:
            case api.Interaction(
                type=api.InteractionType.PING as ixn_type,
                data=None as data
            ):
                callback = None

            case api.Interaction(
                type=api.InteractionType.APPLICATION_COMMAND as ixn_type,
                data=api.ApplicationCommandData(name=command_name) as data
            ):
                callback = self._get_callback_for_command(command_name)

            case api.Interaction(
                type=api.InteractionType.MODAL_SUBMIT as ixn_type,
                data=api.ModalSubmitData(custom_id=modal_id) as data
            ):
                callback = self._get_callback_for_modal(modal_id)

            case api.Interaction(
                type=api.InteractionType.MESSAGE_COMPONENT as ixn_type,
                data=api.MessageComponentData(custom_id=component_id) as data
            ):
                callback = self._get_callback_for_component(component_id)

            case _:
                raise Exception(f"Tried to handle an unknown type of interaction. Interaction is as follows:\nf{pprint.pformat(ixn, compact=True)}")

        response = None
        if isinstance(callback, cb.Callback):
            try:
                response = callback.do(ixn, data)
            except CommandFailure as failure:
                # wrap errors made by users of the bot in an embed
                response = api.ResponseMessage(
                    embeds=[failure.error_embed()],
                    flags=api.MessageFlag.EPHEMERAL
                )
            except Exception:
                # wrap errors made in code in a simple embed, then print the error and continue
                response = api.ResponseMessage(
                    embeds=[CommandFailure(f"When running this command, an exception occurred with the bot code. Let the bot's creator know.").error_embed()],
                    flags=api.MessageFlag.EPHEMERAL
                )
                tb.print_exc()
        return api.InteractionResponse(_match_response_type(response, ixn_type), data=response)
    
    def _get_callback_for_command(self, command_name: str):
        return next(filter(
            lambda on: command_name in on._options,
            (self.on_command, self.on_message, self.on_user)
        ))._options[command_name]
    
    def _get_callback_for_modal(self, modal_id: str):
        return self.on_modal._options[modal_id]
    
    def _get_callback_for_component(self, component_id: str):
        return self.on_component._options[component_id]

    def create_cancel_button(self, ixn_id: api.Snowflake, token: str, in_progress_message: str, button_text: str):
        lock = mp.Event()
        @self.on_component(
            name=ixn_id
        )
        def _():
            lock.set()
            self.delete_response(token)
            self.on_component._options.pop(ixn_id)

        return api.ResponseMessage(
            content=in_progress_message,
            components=[api.ActionRow([
                api.Button(
                    api.ButtonStyle.DANGER,
                    label=button_text,
                    custom_id=ixn_id
                )
            ])]
        ), lock
    
    def send_followup(self, token: str, followup: req.ExecuteWebhook.Form):
        """ Using an interaction message token, sends a followup response after the initial response.
        
            Shouldn't be used before an :attr:`on_command` callback returns / yields. """
        
        return req.ExecuteWebhook(self.id, token, form=followup).do_with(self.token)

    def edit_response(self, token: str, edit: req.EditWebhookMessage.Form, response_id: str="@original"):
        """ Using an interaction message token, edits one of the responses (specified by ID) to that interaction. Defaults to the initial response. """

        return req.EditWebhookMessage(self.id, token, response_id, form=edit).do_with(self.token)

    def delete_response(self, token: str, response_id: str="@original"):
        """ Using an interaction message token, deletes one of the responses (specified by ID) to that interaction. Defaults to the initial response. """

        req.DeleteWebhookMessage(self.id, token, response_id).do_with(self.token)
    
    def send(self, channel_id: api.Snowflake, message: req.CreateMessage.Form):
        """ Sends a standalone message.
        
            The argument ``message`` is a :class:`.req.CreateMessage.Form`. """

        return req.CreateMessage(channel_id, message).do_with(self.token)
    
    def edit(self, channel_id: api.Snowflake, message_id: api.Snowflake, message: req.EditMessage.Form):
        """ Edits a standalone message.
        
            The argument ``message`` is a :class:`.req.EditMessage.Form`. """

        return req.EditMessage(channel_id, message_id, message).do_with(self.token)

    def delete(self, channel_id: api.Snowflake, message_id: api.Snowflake):
        """ Deletes a standalone message. """

        req.DeleteMessage(channel_id, message_id).do_with(self.token)

    def history(self, channel_id: api.Snowflake, limit: int=200, chunk_size: int=50):
        """ Iterates over a channel's messages, starting from the earliest message.
        
            The number of messages retrieved depends on ``limit``. The :class:`.req.GetChannelMessages` request will retrieve a chunk of messages from the channel, the size of which is determined by ``chunk_size``, until the limit is reached. A higher chunk size means less total requests, but those requests will take more time. """

        message: api.Message | None = None
        # limit goes down per message retrieved
        while limit:
            # can't request more than the limit - if the chunk_size would go over the limit, set it instead to the limit
            current_chunk_size = chunk_size if limit > chunk_size else limit
            messages = req.GetChannelMessages(channel_id, req.GetChannelMessages.Query(
                before = message.id if message else None,
                limit = current_chunk_size
            )).do_with(self.token)
            while messages:
                message = messages.pop(0)
                yield message
                limit -= 1
                current_chunk_size -= 1
            if current_chunk_size:
                limit = 0
    
    def send_error(self, channel_id: api.Snowflake, error_text: str):
        """ Sends an error embed with the given text to the given channel. """

        error = CommandFailure(error_text)
        self.send(channel_id, req.CreateMessage.Form(embeds=[error.error_embed()]))

class ChannelSnowflake(api.Snowflake):
    """ This class only serves to differentiate the type of :class:`~.disc.Snowflake` that's stored in a :attr:`.ConfiguredPory.config_model`'s field. Represents the ID of a channel in the guild. """

class RoleSnowflake(api.Snowflake):
    """ This class only serves to differentiate the type of :class`~.disc.Snowflake` that's stored in a :attr:`.ConfiguredPory.config_model`'s field. Represents the ID of a role in the guild. """

t_ConfigType = t.TypeVar("t_ConfigType")
""" Represents the config dataclass used with :class:`.ConfiguredPory`. """

@dc.dataclass
class ConfiguredPory(Pory, t.Generic[t_ConfigType]):
    """ This class adds to :class:`Pory` by wrapping a :attr:`.config_model` dataclass, storing it per guild as json in a given location, and allowing the bot to do things per guild based on that guild's dataclass. A guild's config dataclass instance can be retrieved using :meth:`.get_current_config`. More info on defining a configuration dataclass can be found in :attr:`.config_model`.
    
        It automatically registers a /config slash command that users with Manage Channel permissions can use to configure the dataclass instance for the guild they're in. """

    config_path: str
    """ The path to where the config file will be stored. """
    config_model: type[t_ConfigType]
    """ The dataclass with which to structure the bot.
    
        The dataclass won't recieve any arguments on instantiation - each of its fields must have a default value (or default_factory for fields that store lists).

        :class:`.ConfiguredPory` will leaf through the dataclass's fields to find each field that stores one of the above subclasses of :class:`~.disc.Snowflake`: :class:`.ChannelSnowflake` and :class:`.RoleSnowflake`. It will also find fields that store lists of those snowflake subclasses. For each of the snowflake fields, a subcommand group will be added to the /config slash command that coresponds to that field. Subcommands are then added to that group that will alter that field when used. """

    guild_configs: dict[api.Snowflake, t_ConfigType] = dc.field(default_factory=dict)
    """ The json file, parsed into instances of :attr:`.config_model` per guild. """

    def load(self):
        """ Reads the config file from disk. """
        with open(self.config_path, "r") as f:
            data = json.load(f)
        for guild_id in data:
            self.guild_configs[api.Snowflake(guild_id)] = disc.cast(self.config_model, data[guild_id])
    
    def write(self):
        """ Writes the current config for all guilds to the disk. """
        with open(self.config_path, "w") as f:
            json.dump(disc.uncast(self.guild_configs), f)

    def __post_init__(self, app_id: str):
        super().__post_init__(app_id)

        self.load()

        config = self.on_command.group(
            "config",
            "Configures a channel to do something related to this bot.",
            [api.Permission.MANAGE_CHANNELS]
        )
        
        for field in dc.fields(self.config_model):
            ftype_root = t.get_origin(field.type)
            if ftype_root:
                if not issubclass(ftype_root, (list, types.UnionType)): continue
                if not api.Snowflake in t.get_args(field.type): continue
            elif not field.type is api.Snowflake: continue
            setters = config.group(
                f"{field.name}",
                f"Configure a {'list of channels' if t.get_origin(field.type) else 'channel'} to be related to {field.name}."
            )
            
            # Had to move the function definitions to their own functions!
            # When the functions were defined in this method, the for loop would
            #  iterate past the field the function was meant to be defined for.
            # When the functions were called, they would use the variables in the
            #  scope of the loop's final iteration instead of the variables in the
            #  scope where they were defined.
            # Moving them to a scope that doesn't change when the for-loop changes
            #  fixes this (freezes the field.name, in other words).

            if t.get_origin(field.type) == list:
                setters(self._make_add(field.name))
                setters(self._make_remove(field.name))
            else:
                setters(self._make_set(field.name))
                setters(self._make_unset(field.name))

            
    def get_current_config(self, guild_id: api.Snowflake | None):
        """ Gets the :attr:`.config_model` instance for a specified guild. If ``guild_id`` is ``None``, raises an appropriate :class:`.CommandFailure`.
        
            This method is usually used inside of interaction callbacks. The ``guild_id`` can be retrieved by adding ``guild_id`` to the callback signature as a keyword-specific argument (i.e. an argument that comes after ``*,``). See :func:`~dubious.callback.do_callback` for more details. """

        if guild_id is None:
            raise CommandFailure("You can't use this command outside of a guild.")
        if not self.guild_configs.get(guild_id):
            self.guild_configs[guild_id] = self.config_model()
        return self.guild_configs[guild_id]

    def _make_add(self, name: str):
        def add(channel: api.Channel, *, guild_id: api.Snowflake | None):
            config = self.get_current_config(guild_id)
            channels = getattr(config, name)
            if not isinstance(channels, list):
                raise CommandFailure("Something went wrong: one of the assigned channel lists is a single value instead of a list. Let the bot creator know.")
            channels.append(channel.id)
            self.write()
            return api.ResponseMessage(content=f"Channel <#{channel.id}> added to {name} channels.")
        return add

    def _make_remove(self, name: str):
        def remove(channel: api.Channel, *, guild_id: api.Snowflake | None):
            config = self.get_current_config(guild_id)
            channels = getattr(config, name)
            if not isinstance(channels, list):
                raise CommandFailure("Something went wrong: one of the assigned channel lists is a single value instead of a list. Let the bot creator know.")
            if not channel.id in channels:
                raise CommandFailure(f"That channel isn't a part of the {name} channels.")
            channels.remove(channel.id)
            self.write()
            return api.ResponseMessage(content=f"Channel <#{channel.id}> removed from the {name} channels.")
        return remove
    
    def _make_set(self, name: str):
        def set(channel: api.Channel, *, guild_id: api.Snowflake | None):
            config = self.get_current_config(guild_id)
            setattr(config, name, channel.id)
            self.write()
            return api.ResponseMessage(content=f"Channel <#{channel.id}> has been set as the {name} channel.")
        return set

    def _make_unset(self, name: str):
        def unset(*, guild_id: api.Snowflake | None):
            config = self.get_current_config(guild_id)
            setattr(config, name, None)
            self.write()
            return api.ResponseMessage(content=f"The {name} channel has been set to None.")
        return unset

def get_message_link(guild_id: disc.Snowflake, message: api.Message):
    return f"https://discord.com/channels/{guild_id}/{message.channel_id}/{message.id}"



import dataclasses as dc
import inspect
import multiprocessing as mp
import re
import traceback as tb
import typing as t

from dubious.discord import api

_option_types = {
    str: api.ApplicationCommandOptionType.STRING,
    int: api.ApplicationCommandOptionType.INTEGER,
    bool: api.ApplicationCommandOptionType.BOOLEAN,
    api.User: api.ApplicationCommandOptionType.USER,
    api.Channel: api.ApplicationCommandOptionType.CHANNEL,
    api.Role: api.ApplicationCommandOptionType.ROLE,
    float: api.ApplicationCommandOptionType.NUMBER,
    api.Attachment: api.ApplicationCommandOptionType.ATTACHMENT,
}
_option_types_strs = {
    k.__name__: v for k, v in _option_types.items()
}

_pat_discord_name = re.compile(r"^[-_a-z]{1,32}$")


def _prepare_option(opt: inspect.Parameter):
    if not _pat_discord_name.search(opt.name):
        # https://discord.com/developers/docs/interactions/application-commands#application-command-object-application-command-naming
        raise ValueError(f"Option '{opt.name}' has an invalid name.")
    desc = "No description."
    choices = None
    if t.get_origin(opt.annotation) == t.Annotated:
        anns = t.get_args(opt.annotation)
        typ = _option_types[anns[0]]
        if len(anns) > 1:
            if len(anns[1]) > 100:
                raise ValueError(
                    f"The description for option '{opt.name}' is too long. Should be 100 or less; got {len(anns[1])}.")
            desc = anns[1]
        if len(anns) > 2: choices = anns[2]
    else:
        if isinstance(opt.annotation, str):
            # to account for `from __future__ import annotations` making types strings
            wo_namespace = opt.annotation.split(".")[-1]
            typ = _option_types_strs[wo_namespace]
        else:
            typ = _option_types[opt.annotation]

    return api.ApplicationCommandOption(
        typ,
        opt.name,
        desc,
        required=None if not opt.default == inspect.Parameter.empty else True,
        choices=choices
    )


ta_CommandRet = api.InteractionCallbackData | None | t.Iterator[api.InteractionCallbackData | None]
""" A type alias that defines what a function stored in a :class:`Callback` can return. """

ps_CallbackArgs = t.ParamSpec("ps_CallbackArgs")
""" Represents the arguments for a function stored in a :class:`Callback`. """
t_CommandRet = t.TypeVar("t_CommandRet", bound=ta_CommandRet)
""" Represents the return value for a function stored in a :class:`Callback`. """


def do_callback(
    callback: t.Callable[ps_CallbackArgs, t_CommandRet],
    ixn: api.Interaction,
    data: api.InteractionData | api.ApplicationCommandInteractionDataOption | None
) -> t_CommandRet:
    """ Extracts options from an :class:`.api.Interaction`, and uses them as arguments to call a given ``callback`` function.
    
        This is the "optionated" part of DubiousDiscord: The ``callback`` function's signature can define keyword-specific arguments (i.e. arguments that come after ``*,`` in the signature) that :func:`.do_callback` will take from the ``ixn`` :class:`.api.Interaction` object. Take an example ``callback`` function::
        
            from dubious.discord import api
            from dubious import pory

            bot = pory.Pory("app id", "app key", "bot token")

            @bot.on_command
            def whereami(*, channel_id: api.Snowflake | None):
                if not channel_id:
                    return api.ResponseMessage(content="We're nowhere! How's that possible?")
                else:
                    return api.RepsonseMessage(content=f"We're in the <#{channel_id}> channel!")

        The callback function asks for ``channel_id``, which is a field on :class:`.api.Interaction`. :func:`do_callback` will find the :attr:`~.api.Interaction.channel_id` field on the given ``ixn`` argument and give it to the callback function as a coresponding keyword argument.
        
        Note that the :attr:`.api.Interaction.channel_id` field has the type ``Snowflake | None``. There's no way to enforce this, but the callback function should reflect that type in the signature, as shown in the example. (Though, to be pedantic: In this particular example, the ``whereami`` function will *never* be called in response to a Modal Interaction, which is the only case in which :attr:`~.api.Interaction.channel_id` could be ``None``, so making ``channel_id`` in the signature ``None`` -able isn't as big of a deal.)
        """

    fixed_args: dict[str, t.Any] = {}
    for param in inspect.signature(callback).parameters.values():
        match param.kind:
            case inspect.Parameter.POSITIONAL_OR_KEYWORD:
                assert isinstance(data, (api.ApplicationCommandData, api.ApplicationCommandInteractionDataOption))
                assert isinstance(ixn.data, api.ApplicationCommandData)
                if not param.default == inspect.Parameter.empty: continue
                assert data.options

                found = False
                for opt in data.options:
                    if found := opt.name == param.name:
                        if opt.type == api.ApplicationCommandOptionType.CHANNEL:
                            assert ixn.data.resolved
                            assert isinstance(opt.value, (str, int))
                            assert ixn.data.resolved.channels
                            fixed_args[param.name] = ixn.data.resolved.channels[api.Snowflake(opt.value)]
                        elif opt.type == api.ApplicationCommandOptionType.USER:
                            assert ixn.data.resolved
                            assert isinstance(opt.value, (str, int))
                            assert ixn.data.resolved.users
                            fixed_args[param.name] = ixn.data.resolved.users[api.Snowflake(opt.value)]
                        elif opt.type == api.ApplicationCommandOptionType.ROLE:
                            assert ixn.data.resolved
                            assert isinstance(opt.value, (str, int))
                            assert ixn.data.resolved.roles
                            fixed_args[param.name] = ixn.data.resolved.roles[api.Snowflake(opt.value)]
                        elif opt.type == api.ApplicationCommandOptionType.ATTACHMENT:
                            assert ixn.data.resolved
                            assert isinstance(opt.value, (str, int))
                            assert ixn.data.resolved.attachments
                            fixed_args[param.name] = ixn.data.resolved.attachments[api.Snowflake(opt.value)]
                        else:
                            fixed_args[param.name] = opt.value
                        break
                assert found
            case inspect.Parameter.KEYWORD_ONLY:
                if param.name == "ixn":
                    fixed_args[param.name] = ixn
                elif param.name == "data":
                    fixed_args[param.name] = ixn.data
                elif param.name == "user":
                    if not ixn.user:
                        assert ixn.member
                        fixed_args[param.name] = ixn.member.user
                elif hasattr(ixn, param.name):
                    fixed_args[param.name] = getattr(ixn, param.name)
                elif hasattr(ixn.data, param.name):
                    fixed_args[param.name] = getattr(ixn.data, param.name)
                else:
                    raise Exception()
            case _:
                raise Exception()

    return callback(**fixed_args)  # type: ignore


@dc.dataclass
class Callback(t.Generic[ps_CallbackArgs, t_CommandRet]):
    """ This dataclass holds a function along with an associated name.
    
        The name allows it to be identified dynamically, so that it may be matched with e.g. a Discord interaction. """

    name: str
    __func__: t.Callable[ps_CallbackArgs, t_CommandRet]

    def do(
        self,
        ixn: api.Interaction,
        data: api.InteractionData | api.ApplicationCommandInteractionDataOption | None
    ) -> api.InteractionCallbackData | None:
        """ Calls the wrapped function using data from an :class:`.api.Interaction`.
        
            ``ixn`` and ``data`` are separate arguments because the :attr:`.api.Interaction.data` field can be that of a subcommand group, whose :attr:`~.api.ApplicationCommandData.options` field has only one :class:`.api.ApplicationCommandInteractionDataOption` coresponding to the subcommand (or subcommand group) currently being called. :func:`do_callback` needs to have the :class:`.api.ApplicationCommandInteractionDataOption`\\ s that corespond to the wrapped function's arguments. """

        done = do_callback(self.__func__, ixn, data)
        if isinstance(done, t.Iterator):
            initial_return = next(done)
            mp.Process(target=list, args=(done,)).start()
            return initial_return
        return done


@dc.dataclass
class CallbackGroup:
    _options: dict[str, Callback] = dc.field(default_factory=dict, kw_only=True)

    def __call__(self, cb: t.Callable[ps_CallbackArgs, t_CommandRet] | None = None, /, *, name: str | None = None):
        def _(callback: t.Callable[ps_CallbackArgs, t_CommandRet]):
            _name = name if name else callback.__name__
            self._options[_name] = Callback(_name, callback)
            return callback

        if cb:
            return _(cb)
        else:
            return _


@dc.dataclass
class Command(api.ApplicationCommandOption, Callback[ps_CallbackArgs, t_CommandRet]):
    """ This class contains a callback function to be matched against a Discord Interaction. It subclasses :class:`.ApplicationCommandOption` to vaguely match the structure of incoming commands.
        
        Generally, this class isn't meant to be instantiated manually. Instead, the right way to create an instance is to use a :class:`CommandGroup` as a decorator around a callback function. """

    __func__: t.Callable[ps_CallbackArgs, t_CommandRet]

    default_member_permissions: str | None = None
    """ Defines the permissions required to use the command. This field comes from the :class:`.api.ApplicationCommand` class.
    
        The API documentation marks it as a string, even though it's mean to be a bitfield of :class:`.api.Permission` bits. """
    
    _options: dict[str, api.ApplicationCommandOption] = dc.field(default_factory=dict, kw_only=True)

    guild_id: api.Snowflake | None = None
    """ Defining this can specify a guild in which to register the command to Discord. """

    early_response: t_CommandRet | None = None

    @property
    def options(self):
        return list(self._options.values()) if self._options else None

    @options.setter
    def options(self, _: list[api.ApplicationCommandOption] | None):
        # only needed to appease dataclass default __init__
        pass

    def __post_init__(self):
        assert _pat_discord_name.search(self.name), f"Command '{self.name}' has an invalid name."
        assert len(
            self.description) <= 100, f"The description on command '{self.name}' is too long. Should be 100 or less; got {len(self.description)}."

        for opt in inspect.signature(self.__func__).parameters.values():
            if opt.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                try:
                    self._options[opt.name] = _prepare_option(opt)
                except Exception as e:
                    raise ValueError(f"In command '{self.name}': {''.join(tb.format_exception(e))}")
        if len(self._options) > 25: raise ValueError(
            f"There are too many options for command \"{self.name}\". Should be 25 or less; got {len(self._options)}.")

    def _compare_with(self, registered: api.ApplicationCommand):
        return (
            self.guild_id == registered.guild_id and
            self.default_member_permissions == registered.default_member_permissions and
            self._compare_with_as_option(registered)
        )

    def _compare_with_as_option(self, registered: api.ApplicationCommand | api.ApplicationCommandOption):
        return (
            self.name == registered.name and
            self.name_localizations == registered.name_localizations and
            self.description == registered.description and
            self.description_localizations == registered.description_localizations and
            all(
                this_opt._compare_with_as_option(that_opt)
                if isinstance(this_opt, Command) else
                this_opt == that_opt
                for this_opt, that_opt in zip(self.options, registered.options)
            ) if self.options and registered.options else self.options == registered.options
        )

    def __call__(self, *args: ps_CallbackArgs.args, **kwargs: ps_CallbackArgs.kwargs):
        """ Redefines what calling an instance of this class does to be the same as calling the wrapped function. """

        return self.__func__(*args, **kwargs)


@dc.dataclass
class CommandGroup(CallbackGroup, Command[[], None]):
    _options: dict[str, Command] = dc.field(default_factory=dict, kw_only=True)

    def get_commands(self, *, as_type: api.ApplicationCommandType | None = None):
        prepared_commands: dict[str, Command] = {}
        for opt in self._options.values():
            prepared = opt
            if as_type:
                # below should type error but dc.replace doesn't care
                # api.ApplicationCommandOption.type field can't be type api.ApplicationCommandType
                prepared = dc.replace(prepared, type=as_type)
            if isinstance(prepared, CommandGroup):
                prepared = dc.replace(prepared, _options=prepared.get_commands())
            prepared_commands[opt.name] = prepared
        return prepared_commands

    @classmethod
    def new(cls):
        return cls(
            "_",
            lambda: None,
            api.ApplicationCommandOptionType.SUB_COMMAND_GROUP,
            "_",
        )

    def __call__(self, cb: t.Callable[ps_CallbackArgs, t_CommandRet] | None = None, /, *,
                 name: str | None = None,
                 desc: str | None = None,
                 perms: list[api.Permission] | None = None,
                 guild: api.Snowflake | None = None,
                 ):
        def _wrap(callback: t.Callable[ps_CallbackArgs, t_CommandRet]) -> t.Callable[ps_CallbackArgs, t_CommandRet]:
            _name = name if name else callback.__name__
            _desc = desc if desc else callback.__doc__ if (not desc) and callback.__doc__ else "No description provided."
            _perms = 0
            for perm in perms if perms else []:
                _perms |= perm
            self._options[_name] = Command(
                _name,
                callback,
                api.ApplicationCommandOptionType.SUB_COMMAND,
                _desc.strip(),
                str(_perms) if perms else None,
                guild,
            )
            return callback

        return _wrap(cb) if cb else _wrap

    def group(self,
              name: str,
              desc: str,
              perms: t.Collection[api.Permission] | None = None,
              guild: api.Snowflake | None = None
              ):
        """ Adds a subcommand :class:`CommandGroup` to this :class:`CommandGroup`'s options and returns it. """

        _perms = 0
        if perms:
            for perm in perms:
                _perms |= perm
        group = self.__class__(
            name,
            lambda: None,
            api.ApplicationCommandOptionType.SUB_COMMAND_GROUP,
            desc,
            str(_perms) if perms else None,
            guild
        )
        self._options[name] = group
        return group
    
    def do(self, ixn: api.Interaction, data: api.ApplicationCommandData | api.ApplicationCommandInteractionDataOption) -> api.InteractionCallbackData | None:
        """ Propagates an :class:`.api.Interaction` down to the subcommand callback it's meant for.
        
            In other words, this function extracts the appropriate subcommand :class:`Command` object and calls the subcommand's :meth:`Callback.do`. If the subcommand is a :class:`CommandGroup`, this function recurses by calling the subcommand's :meth:`CommandGroup.do`. """

        assert data.options
        # data.options is always guaranteed to have a length of 1 as long as the data belongs to a subcommand group
        option, = data.options
        subcommand = self._options[option.name]
        return subcommand.do(ixn, option)




from __future__ import annotations

import abc
import dataclasses as dc
import enum
import math
import time
import types
import typing as t

from requests import request


class Snowflake(str):
    """ Represents a `Discord Snowflake <https://discord.com/developers/docs/reference#snowflakes>`_.
    
        :mod:`dubious.discord.api` imports this. It's acceptable to ``from dubious.discord import api`` and use ``api.Snowflake`` instead of ``disc.Snowflake``. """

    def __init__(self, r: str|int):
        self.id = int(r) if isinstance(r, str) else r
        self.timestamp = (self.id >> 22) + 1420070400000
        self.workerID = (self.id & 0x3E0000) >> 17
        self.processID = (self.id & 0x1F000) >> 12
        self.increment = self.id & 0xFFF

    def __repr__(self):
        return str(self.id)

    def __str__(self) -> str:
        return repr(self)

    def __hash__(self):
        return self.id

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            try:
                return int(o) == self.id # type: ignore
            except TypeError:
                return False
        return o.id == self.id

    def __ne__(self, o: object) -> bool:
        return not self == o

@dc.dataclass
class Disc:
    """ Root class for all Discord objects. Exists in case there's some functionality to be implemented across every Discord data structure. """

t_Cast = t.TypeVar("t_Cast")
""" Represents the return type of :func:`.cast`."""
def cast(t_to: type[t_Cast], raw: t.Any) -> t_Cast:
    """ Casts ``raw`` data of any shape to the given type, ``t_to``.  """

    if raw is None: return raw

    if dc.is_dataclass(raw):
        raw = dc.asdict(raw)
    if dc.is_dataclass(t_to):
        fixedraw = {}

        fieldtypes: dict[str, type] = t.get_type_hints(t_to)
        fixedraw = {
            field.name: cast(
                fieldtypes[field.name], raw[field.name] if field.name in raw else None
            ) for field in dc.fields(t_to)
        }
        return t_to(**fixedraw)

    t_root = t.get_origin(t_to)

    if t_root:
        if issubclass(t_root, list):
            t_list, *_ = t.get_args(t_to)
            return [
                cast(t_list, rawitem) for rawitem in raw
            ] # type: ignore
        elif issubclass(t_root, dict):
            t_key, t_val, *_ = t.get_args(t_to)
            return {
                cast(t_key, key): cast(t_val, val)
                    for key, val in raw.items()
            } # type: ignore
        elif issubclass(t_root, types.UnionType):
            t_optionals = [
                t for t in t.get_args(t_to)
                    if t != type(None)
            ]
            best_score = 0
            best_len = math.inf
            best_opt = None
            for t_opt in t_optionals:
                if dc.is_dataclass(t_opt):
                    fields = [field.name for field in dc.fields(t_opt)]
                    score = sum(name in fields for name in raw)
                    if (
                        score > best_score or (
                            score == best_score and 
                            len(fields) < best_len
                        )
                    ):
                        best_score = score
                        best_len = len(fields)
                        best_opt = t_opt
                else:
                    try:
                        return cast(t_opt, raw)
                    except ValueError:
                        continue
            return cast(best_opt, raw) #type: ignore

    try:
        return t_to(raw)
    except ValueError:
        return raw

def uncast(raw: t.Any):
    """ Transforms an object to a json-compatible object. """

    if dc.is_dataclass(raw):
        return dc.asdict(raw)
    elif isinstance(raw, dict):
        return {uncast(k): uncast(v) for k, v in raw.items()}
    elif isinstance(raw, list):
        return [uncast(v) for v in raw]
    else:
        return raw

ROOT = "https://discord.com/api"

class Http(str, enum.Enum):
    """ Enum for each potential type of request to make to the Discord API. """

    GET = "GET"
    PUT = "PUT"
    PATCH = "PATCH"
    POST = "POST"
    DELETE = "DELETE"

t_Ret = t.TypeVar("t_Ret")
""" Represents the type that an :class:`HttpReq` will return upon success. """
class HttpReq(abc.ABC, t.Generic[t_Ret]):
    """ A class that holds information about a request to make to the Discord API.
     
        Subclasses should define within themselves a class for the :attr:`.query` attribute and/or the :attr:`.form` attribute, if either exist, and appropriately set the type hints of either. """
    
    query: Disc | None = None
    """ An object to give to the request. Acts as a query string in the URL; is given to :func:`requests.request` as the ``params`` kwarg. """
    form: Disc | None = None
    """ An object to give to the request. Acts as a json string in the request's body; is given to :func:`requests.request` as the ``json`` kwarg. """

    method: t.ClassVar[Http]
    """ The HTTP method with which to make the request. Given to :func:`requests.request` as the ``method`` argument. """
    endpoint: str
    """ The URL at which the request should be made. Given to :func:`requests.request` as the ``url`` argument. This field is set dynamically in a subclass's ``__post_init__`` in order to allow for per-guild/per-channel/etc. requests. """

    @abc.abstractmethod
    def cast(self, data: t.Any) -> t_Ret:
        ...

    def do_with(self, token: str) -> t_Ret:
        """ Perform this Discord API request with :func:`requests.request`. Requires authentication using a bot token - this is usually :attr:`dubious.pory.Pory.token`.
        
            When a rate limit is encountered, wait the response's given time and retry (by calling recursively). """

        res = request(self.method, ROOT + self.endpoint,
            headers={
                "Authorization": f"Bot {token}"
            },
            params=dc.asdict(self.query) if self.query else None,
            json=dc.asdict(self.form) if self.form else None
        )
        if not res.status_code in range(200, 300):
            error = res.json()
            if error.get("retry_after"):
                wait = error["retry_after"]/1000
                print(f"rate limited, waiting {wait} seconds")
                time.sleep(wait)
                return self.do_with(token)
            raise Exception(f"{res.status_code}: {str(error)}")
        return self.cast(res.json() if res.text else None)




""" The :mod:`dubious.discord` module contains the dataclasses for the Discord API's objects.

    * :mod:`.disc` is made to deal with transforming Discord objects to and from json.
    * :mod:`.api` defines the structures of all of the objects that Discord will send in Interactions.
    * :mod:`.req` defines the structures of all of the objects that Discord accepts/sends back from http requests to its API.
    
    """

from dubious.discord import disc, api, req



import os
import sys
import sphinx.errors
sys.path.insert(0, os.path.abspath("../src"))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'DubiousDiscord'
copyright = '2023, Carl Best'
author = 'lapras'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc", "sphinx.ext.intersphinx"]

autodoc_default_options = {
    "member-order": "bysource",
}
autodoc_type_aliases = {
    "ps_CallbackArgs": "ps_CallbackArgs"
}
autodoc_typehints_format = "short"

def missing_reference(_app, _domain, node, _contnode):
    print(node["reftarget"])
    if any([ignore in node["reftarget"] for ignore in ["ClassVar", "InitVar", "ps_CallbackArgs"]]):
        raise sphinx.errors.NoUri()

def setup(app):
    app.connect("missing-reference", missing_reference)

nitpicky = True

python_display_short_literal_types = True
python_use_unqualified_type_names = True
maximum_signature_line_length = 200

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    "flask": ("https://flask.palletsprojects.com/en/2.3.x", None),
    "requests": ("https://requests.readthedocs.io/en/latest", None)
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output




from __future__ import annotations
import abc
import dataclasses as dc
import enum
from pprint import pprint
import re
import typing as t

eval_name = re.compile(r"\w")

ta_MatchResults = dict[str, "str | list[str] | ta_MatchResults"]
ta_Process = t.Callable[[str, "Trick"], ta_MatchResults]
ta_Instructions = dict[str, list[ta_Process]]

@dc.dataclass
class Step:
    name: str | None = None
    ops: list[str] = dc.field(default_factory=list)
    opname: str | None = None
    inner: str = ""
    nest: int = 0
    done: bool = False

@dc.dataclass
class Trick:
    template: str
    instructions: ta_Instructions = dc.field(default_factory=dict)
    pattern: str = dc.field(init=False, default="")

    def __post_init__(self):
        print(f"trick: {self.template}")
        step = Step()
        for char in self.template:
            print(char)
            match char, step:
                case "$", Step(name=None):
                    print("var start")
                    step.name = ""
                case _, Step(name=None):
                    print("adding to pattern")
                    self.pattern += char
                case _, Step(name=str(some), opname=None, nest=0) if re.match(r"\w", char):
                    print("adding to var name")
                    step.name = some + char
                case ".", Step(name=str(_), opname=None, nest=0):
                    print("op start")
                    step.opname = ""
                case _, Step(name=str(some), opname=None, nest=0) if not re.match(r"\w", char):
                    print("var end")
                    self.create_group(some, [])
                    step.name = None
                case _, Step(name=str(_), opname=str(some), nest=0) if re.match(r"\w", char):
                    print("adding to op")
                    step.opname = some + char
                case "{", Step(name=str(_), opname=str(some), nest=0):
                    print("up nest (new op)")
                    step.ops.append(some)
                    step.opname = None
                    step.nest += 1
                case "{", Step(name=str(_), nest=int(some)) if some > 0:
                    print("up nest")
                    step.nest += 1
                case "}", Step(name=str(name), nest=int(some)) if some > 0:
                    print("down nest")
                    step.nest -= 1
                    if step.nest == 0:
                        #self.create_group(name, [op_by_char[opname] for opname in step.ops], step.inner)
                        step.name = None
                case _, Step(name=str(_), nest=int(some)) if some > 0:
                    print("adding to inner")
                    step.inner += char
                case _:
                    raise Exception()
        print(f"trick [[{self.pattern}]] done")

    def create_group(self, name: str, instructions: list[ta_Process], inner: str=""):
        self.pattern += r"(?P<"+name+r">[\w\W]+?)"
        #self.instructions[name] = Trick(inner, instructions) if inner else instructions

    def match(self, on: str):
        match = re.match(self.pattern, on)
        if not match: return
        groups = match.groupdict()
        results: ta_MatchResults = {}
        for group in groups:
            inner = self.instructions[group]
            got = groups[group]
            if not inner:
                results[group] = got
            # else:
            #     results[group] = inner.match(got)

"""
Goal:
    Create a way to store human-readable enumerable information in Discord messages, and create a way to parse that information back out of those messages.
    Examples:
        "Query: `$query_tokens.sep{, }`"
        - "Query: `":
            Literal; doesn't store information.
        - "$query_tokens":
            This will be replaced with a variable called "query_tokens", given when serializing.
            The intent is for this variable to be a list of strings.
        - ".sep{, }":
            This modifier, appended to the query_tokens variable, will separate the variable's items by commas.
        - "`":
            Literal.
        
        "$allowed_users.empty?.lit{Anyone}.else.sep{, }.fmt{<@{}>} can use this query."
        - "$allowed_users":
            Use a variable named "allowed_users".
        - ".empty?":
            This modifier, appended to the allowed_users variable, will only include the modifiers following it up until the ".else" modifier if the given value for the variable is empty.
        - ".lit{Anyone}":
            This modifier will evaluate to the literal string "Anyone".
        - ".else":
            This modifier ends the modifiers for a conditional modifier and begins the modifiers for the alternative.
        - ".sep{, }":
            Separates the variable's items by commas.
        - ".fmt{<@{}>}":
            Surrounds each of the variable's items with <@ and >, replacing the {} with the item itself.
        - " can use this query.":
            Literal.
    
    Can't use regular expressions to parse the rules - the final character of a variable and its modifiers is whitespace, but there can be whitespace inside of curly brackets without ending the modifiers.
    Conditional modifiers can't be nested - modifiers can't be nested inside of other modifiers.
"""

class States(enum.Enum):
    Root = 0
    Variable = 1
    Modifier = 2
    ModifierArg = 3
    EndingVariable = 4

@dc.dataclass
class Modifier:
    name: str = ""
    arg: str = ""

@dc.dataclass
class Variable:
    name: str = ""
    modifiers: str = ""

def parse_modifiers(text: str):
    modifiers: list[Modifier] = []
    state = States.Root
    nest_level = 0
    for char in text:
        match state, char:
            case States.Root | States.Modifier | States.EndingVariable, ".":
                state = States.Modifier
                modifiers.append(Modifier())
            case States.Modifier, "{":
                state = States.ModifierArg
                nest_level += 1
            case States.Modifier, char:
                modifiers[-1].name += char

            case States.ModifierArg, char:
                if char == "{":
                    nest_level += 1
                elif char == "}":
                    nest_level -= 1
                if nest_level == 0:
                    state = States.EndingVariable
                else:
                    modifiers[-1].arg += char
    return [modifier_methods[modifier.name](modifier.arg) for modifier in modifiers]

def parse_variables(text: str):
    state = States.Root
    variables: list[Variable] = []
    regex = ""
    nest_level = 0
    for char in text:
        match state, char:
            case States.Root, "$":
                variables.append(Variable())
                state = States.Variable
            case States.Root, char:
                regex += char
            case States.Variable | States.Modifier | States.EndingVariable, ".":
                if state == States.Variable:
                    regex += r"(?P<"+variables[-1].name+r">[\w\W]+?)"
                variables[-1].modifiers += char
                state = States.Modifier
            case States.Variable, char:
                variables[-1].name += char
            case States.Modifier, char:
                variables[-1].modifiers += char
                if char == "{":
                    nest_level += 1
                elif char == "}":
                    nest_level -= 1
                    if nest_level == 0:
                        state = States.EndingVariable
            case States.EndingVariable, char:
                regex += char
                state = States.Root
    return variables, regex

def decode(modifiers: list[ModifierMethods], using: str):
    decoded: list[str] | str = using
    for mod in reversed(modifiers):
        decoded = mod.dec(decoded)
    if not isinstance(decoded, list):
        raise Exception(f"When processing decoding modifiers, the final modifier returned a single string instead of a list.")
    return decoded

def encode(modifiers: list[ModifierMethods], using: list[str]):
    encoded: list[str] | str = [*using]
    for mod in modifiers:
        encoded = mod.enc(encoded)
    if not isinstance(encoded, str):
        raise Exception(f"When processing encoding modifiers, the final modifier returned a list instead of a single string.")
    return encoded

@dc.dataclass
class Rules:
    variable_modifiers: dict[str, list[ModifierMethods]] = dc.field(init=False, default_factory=dict)
    regex: re.Pattern = dc.field(init=False)
    raw: str = dc.field(init=False)

    rules_text: dc.InitVar[str]

    def __post_init__(self, rules_text: str):
        variables, regex = parse_variables(rules_text)
        print(variables, regex)
        
        for variable in variables:
            self.variable_modifiers[variable.name] = parse_modifiers(variable.modifiers)

        self.regex = re.compile(regex)
        self.raw = regex
    
    def decode(self, against: str):
        match = self.regex.match(against)
        if not match: return {}
        groups = match.groupdict()
        modified: dict[str, list[str]] = {}
        for var_name, encoded in groups.items():
            modified[var_name] = decode(self.variable_modifiers[var_name], encoded)
        return modified
    
    def encode(self, **variables: list[str]):
        template = self.raw
        modified: dict[str, str] = {}
        for variable_name in variables:
            modified[variable_name] = encode(self.variable_modifiers[variable_name], variables[variable_name])
        for modified_var_name, encoded in modified.items():
            template = template.replace(r"(?P<"+modified_var_name+r">[\w\W]+?)", encoded)
        return template

modifier_methods: dict[str, t.Type[ModifierMethods]] = {}
def modifier(name: str):
    def _(cls: t.Type[ModifierMethods]):
        modifier_methods[name] = cls
    return _

@dc.dataclass
class ModifierMethods(abc.ABC):
    arg: str

    @abc.abstractmethod
    def enc(self, to_encode: str | list[str]) -> str | list[str]: ...

    @abc.abstractmethod
    def dec(self, to_decode: str | list[str]) -> str | list[str]: ...

@modifier("sep")
@dc.dataclass
class Separate(ModifierMethods):
    def enc(self, to_encode: list[str]):
        return self.arg.join(to_encode)
    def dec(self, to_decode: str):
        return to_decode.split(self.arg)

@modifier("fmt")
@dc.dataclass
class Format(ModifierMethods):
    def enc(self, to_encode: list[str]):
        return [re.sub(r"{}", item, self.arg) for item in to_encode]
    def dec(self, to_decode: list[str]):
        lstrip, rstrip = self.arg.split(r"{}")
        return [item.lstrip(lstrip).rstrip(rstrip) for item in to_decode]

@modifier("empty?")
@dc.dataclass
class Empty(ModifierMethods):
    literal: str = dc.field(init=False)
    further_modifiers: list[ModifierMethods] = dc.field(init=False)
    def __post_init__(self):
        self.literal, further_modifiers = self.arg.split(",", 1)
        self.further_modifiers = parse_modifiers(further_modifiers)

    def enc(self, to_encode: list[str]):
        return self.literal if not len(to_encode) else encode(self.further_modifiers, to_encode)
    def dec(self, to_decode: str):
        return [] if to_decode == self.literal else decode(self.further_modifiers, to_decode)

if __name__ == "__main__":
    trick1 = Rules(r"Query: `$query_tokens.sep{, }`")
    trick2 = Rules(r"$allowed_users.empty?{Anyone, .fmt{<@{}>}.sep{, }} can use this query.")
    pprint(trick1)
    pprint(trick2)
    t1_dec = trick1.decode("Query: `1, 2, 3, 4`")
    t2_dec = trick2.decode("Anyone can use this query.")
    t2_dec2 = trick2.decode("<@lapras#0594> can use this query.")
    print(t1_dec, t2_dec, t2_dec2)
    t1_enc = trick1.encode(**t1_dec)
    t2_enc = trick2.encode(**t2_dec)
    t2_enc2 = trick2.encode(**t2_dec2)
    print(t1_enc, t2_enc, t2_enc2)




""" DubiousDiscord is an opinionated Discord API wrapper written in Python.

    It features:

    * a full set of dataclasses automatically generated by parsing the `Discord API Documentation <https://github.com/discord/discord-api-docs>`_,
    * a decorator-based `Interaction <https://discord.com/developers/docs/interactions/application-commands>`_ builder through :class:`.pory.Pory`, and
    * a small DSL through :mod:`~dubious.zap` to make transforming information into and out of human-readable text (mostly Discord messages) easier.
    """

from dubious import pory, callback as cb, zap